# Copyright 2008 Matt Chaput. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#    1. Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#
#    2. Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY MATT CHAPUT ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
# EVENT SHALL MATT CHAPUT OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
# EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# The views and conclusions contained in the software and documentation are
# those of the authors and should not be interpreted as representing official
# policies, either expressed or implied, of Matt Chaput.

"""The highlight module contains classes and functions for displaying short
excerpts from hit documents in the search results you present to the user, with
query terms highlighted.
"""

from __future__ import division
from collections import deque
from heapq import nlargest
from cgi import escape as htmlescape


# Fragment object

class Fragment(object):
    """Represents a fragment (extract) from a hit document. This object is
    mainly used to keep track of the start and end points of the fragment; it
    does not contain the text of the fragment or do much else.
    """
    
    def __init__(self, tokens, charsbefore=0, charsafter=0, textlen=999999):
        """
        :param tokens: list of the Token objects in the fragment. 
        :param charsbefore: approx. how many characters before the start of the
            first matched term to include in the fragment.
        :param charsafter: approx. how many characters after the end of the
            last matched term to include in the fragment.
        :param textlen: length in characters of the document text.
        """
        
        #: index of the first character of the fragment in the original
        # document
        self.startchar = max(0, tokens[0].startchar - charsbefore)
        #: index after the last character of the fragment in the original
        #document
        self.endchar = min(textlen, tokens[-1].endchar + charsafter)
        self.matches = [t for t in tokens if t.matched]
        self.matched_terms = frozenset(t.text for t in self.matches)
    
    def __len__(self):
        return self.endchar - self.startchar
    
    def overlaps(self, fragment):
        sc = self.startchar
        ec = self.endchar
        fsc = fragment.startchar
        fec = fragment.endchar
        return (fsc > sc and fsc < ec) or (fec > sc and fec < ec)
    
    def overlapped_length(self, fragment):
        sc = self.startchar
        ec = self.endchar
        fsc = fragment.startchar
        fec = fragment.endchar
        return max(ec, fec) - min(sc, fsc)
    
    def has_matches(self):
        return any(t.matched for t in self.tokens)
    

# Filters

def copyandmatchfilter(termset, tokens):
    for t in tokens:
        t = t.copy()
        t.matched = t.text in termset
        yield t


# Fragmenters

class Fragmenter(object):
    def __call__(self, text, tokens):
        raise NotImplementedError
    

class WholeFragmenter(Fragmenter):
    def __call__(self, text, tokens):
        """Doesn't fragment the token stream. This object just returns the
        entire stream as one "fragment". This is useful if you want to
        highlight the entire text.
        """
        
        tokens = list(tokens)
        before = after = 0
        if tokens:
            before = tokens[0].startchar
            after = len(text) - tokens[-1].endchar
        return [Fragment(tokens, charsbefore=before, charsafter=after)]


# Backwards compatiblity
NullFragmeter = WholeFragmenter


class SimpleFragmenter(Fragmenter):
    """Simply splits the text into roughly equal sized chunks.
    """
    
    def __init__(self, size=70):
        """
        :param size: size (in characters) to chunk to. The chunking is based on
            tokens, so the fragments will usually be smaller.
        """
        self.size = size
    
    def __call__(self, text, tokens):
        size = self.size
        first = None
        frag = []
        
        for t in tokens:
            if first is None:
                first = t.startchar
            
            if t.endchar - first > size:
                first = None
                if frag:
                    yield Fragment(frag)
                frag = []
            
            frag.append(t)
            
        if frag:
            yield Fragment(frag)


class SentenceFragmenter(Fragmenter):
    """Breaks the text up on sentence end punctuation characters
    (".", "!", or "?"). This object works by looking in the original text for a
    sentence end as the next character after each token's 'endchar'.
    
    When highlighting with this fragmenter, you should use an analyzer that
    does NOT remove stop words, for example::
    
        sa = StandardAnalyzer(stoplist=None)
    """
    
    def __init__(self, maxchars=200, sentencechars=".!?"):
        """
        :param maxchars: The maximum number of characters allowed in a fragment.
        """
        
        self.maxchars = maxchars
        self.sentencechars = frozenset(sentencechars)
    
    def __call__(self, text, tokens):
        maxchars = self.maxchars
        sentencechars = self.sentencechars
        textlen = len(text)
        first = None
        frag = []
        
        for t in tokens:
            if first is None:
                first = t.startchar
            endchar = t.endchar
            
            if endchar - first > maxchars:
                first = None
                if frag:
                    yield Fragment(frag)
                frag = []
            
            frag.append(t)
            if frag and endchar < textlen and text[endchar] in sentencechars:
                # Don't break for two periods in a row (e.g. ignore "...")
                if endchar + 1 < textlen and text[endchar + 1] in sentencechars:
                    continue
                
                yield Fragment(frag, charsafter=0)
                frag = []
                first = None
        
        if frag:
            yield Fragment(frag)


class ContextFragmenter(Fragmenter):
    """Looks for matched terms and aggregates them with their surrounding
    context.
    
    This fragmenter only yields fragments that contain matched terms.
    """
    
    def __init__(self, maxchars=200, surround=20):
        """
        :param maxchars: The maximum number of characters allowed in a
            fragment.
        :param surround: The number of extra characters of context to add both
            before the first matched term and after the last matched term.
        """
        
        self.maxchars = maxchars
        self.charsbefore = self.charsafter = surround
    
    def __call__(self, text, tokens):
        maxchars = self.maxchars
        charsbefore = self.charsbefore
        charsafter = self.charsafter
        
        current = deque()
        currentlen = 0
        countdown = -1
        for t in tokens:
            if t.matched:
                countdown = charsafter
                # Add on "unused" context length from the front
                countdown += (charsbefore - currentlen)
            
            current.append(t)
            
            length = t.endchar - t.startchar
            currentlen += length
            
            if countdown >= 0:
                countdown -= length
                
                if countdown < 0 or currentlen >= maxchars:
                    yield Fragment(current)
                    current = deque()
                    currentlen = 0
            
            else:
                while current and currentlen > charsbefore:
                    t = current.popleft()
                    currentlen -= t.endchar - t.startchar

        if countdown >= 0:
            yield Fragment(current)


#class VectorFragmenter(Fragmenter):
#    def __init__(self, termmap, maxchars=200, charsbefore=20, charsafter=20):
#        """
#        :param termmap: A dictionary mapping the terms you're looking for to
#            lists of either (posn, startchar, endchar) or
#            (posn, startchar, endchar, boost) tuples.
#        :param maxchars: The maximum number of characters allowed in a fragment.
#        :param charsbefore: The number of extra characters of context to add before
#            the first matched term.
#        :param charsafter: The number of extra characters of context to add after
#            the last matched term.
#        """
#        
#        self.termmap = termmap
#        self.maxchars = maxchars
#        self.charsbefore = charsbefore
#        self.charsafter = charsafter
#    
#    def __call__(self, text, tokens):
#        maxchars = self.maxchars
#        charsbefore = self.charsbefore
#        charsafter = self.charsafter
#        textlen = len(text)
#        
#        vfrags = []
#        for term, data in self.termmap.iteritems():
#            if len(data) == 3:
#                t = Token(startchar = data[1], endchar = data[2])
#            elif len(data) == 4:
#                t = Token(startchar = data[1], endchar = data[2], boost = data[3])
#            else:
#                raise ValueError(repr(data))
#            
#            newfrag = VFragment([t], charsbefore, charsafter, textlen)
#            added = False
#            
#            for vf in vfrags:
#                if vf.overlaps(newfrag) and vf.overlapped_length(newfrag) < maxchars:
#                    vf.merge(newfrag)
#                    added = True
#                    break


# Fragment scorers

class FragmentScorer(object):
    pass


class BasicFragmentScorer(FragmentScorer):
    def __call__(self, f):
        # Add up the boosts for the matched terms in this passage
        score = sum(t.boost for t in f.matches)
        
        # Favor diversity: multiply score by the number of separate
        # terms matched
        score *= len(f.matched_terms) * 100
        
        return score


# Fragment sorters

def SCORE(fragment):
    "Sorts higher scored passages first."
    return None


def FIRST(fragment):
    "Sorts passages from earlier in the document first."
    return fragment.startchar


def LONGER(fragment):
    "Sorts longer passages first."
    return 0 - len(fragment)


def SHORTER(fragment):
    "Sort shorter passages first."
    return len(fragment)


# Formatters

class Formatter(object):
    pass


class UppercaseFormatter(Formatter):
    """Returns a string in which the matched terms are in UPPERCASE.
    """
    
    def __init__(self, between="..."):
        """
        :param between: the text to add between fragments.
        """
        
        self.between = between
        
    def _format_fragment(self, text, fragment):
        output = []
        index = fragment.startchar
        
        for t in fragment.matches:
            if t.startchar > index:
                output.append(text[index:t.startchar])
            
            ttxt = text[t.startchar:t.endchar]
            if t.matched:
                ttxt = ttxt.upper()
            output.append(ttxt)
            index = t.endchar
        
        output.append(text[index:fragment.endchar])
        return "".join(output)

    def __call__(self, text, fragments):
        return self.between.join((self._format_fragment(text, fragment)
                                  for fragment in fragments))


class HtmlFormatter(Formatter):
    """Returns a string containing HTML formatting around the matched terms.
    
    This formatter wraps matched terms in an HTML element with two class names.
    The first class name (set with the constructor argument ``classname``) is
    the same for each match. The second class name (set with the constructor
    argument ``termclass`` is different depending on which term matched. This
    allows you to give different formatting (for example, different background
    colors) to the different terms in the excerpt.
    
    >>> hf = HtmlFormatter(tagname="span", classname="match", termclass="term")
    >>> hf(mytext, myfragments)
    "The <span class="match term0">template</span> <span class="match term1">geometry</span> is..."
    
    This object maintains a dictionary mapping terms to HTML class names (e.g.
    ``term0`` and ``term1`` above), so that multiple excerpts will use the same
    class for the same term. If you want to re-use the same HtmlFormatter
    object with different searches, you should call HtmlFormatter.clear()
    between searches to clear the mapping.
    """
    
    template = '<%(tag)s class=%(q)s%(cls)s%(tn)s%(q)s>%(t)s</%(tag)s>'
    
    def __init__(self, tagname="strong", between="...",
                 classname="match", termclass="term", maxclasses=5,
                 attrquote='"'):
        """
        :param tagname: the tag to wrap around matching terms.
        :param between: the text to add between fragments.
        :param classname: the class name to add to the elements wrapped around
            matching terms.
        :param termclass: the class name prefix for the second class which is
            different for each matched term.
        :param maxclasses: the maximum number of term classes to produce. This
            limits the number of classes you have to define in CSS by recycling
            term class names. For example, if you set maxclasses to 3 and have
            5 terms, the 5 terms will use the CSS classes ``term0``, ``term1``,
            ``term2``, ``term0``, ``term1``.
        """
        
        self.between = between
        self.tagname = tagname
        self.classname = classname
        self.termclass = termclass
        self.attrquote = attrquote
        self.maxclasses = maxclasses
        self.seen = {}
    
    def _format_fragment(self, text, fragment, seen):
        htmlclass = " ".join((self.classname, self.termclass))
        
        output = []
        index = fragment.startchar
        
        for t in fragment.matches:
            if t.startchar > index:
                output.append(text[index:t.startchar])
            
            ttxt = htmlescape(text[t.startchar:t.endchar])
            if t.matched:
                if t.text in seen:
                    termnum = seen[t.text]
                else:
                    termnum = len(seen) % self.maxclasses
                    seen[t.text] = termnum
                ttxt = self.template % {"tag": self.tagname,
                                        "q": self.attrquote,
                                        "cls": htmlclass,
                                        "t": ttxt, "tn": termnum}
            output.append(ttxt)
            index = t.endchar
        
        if index < fragment.endchar:
            output.append(text[index:fragment.endchar])
        
        return "".join(output)
    
    def __call__(self, text, fragments):
        seen = self.seen
        return self.between.join(self._format_fragment(text, fragment, seen)
                                 for fragment in fragments)
    
    def clean(self):
        """Clears the dictionary mapping terms to HTML classnames.
        """
        self.seen = {}


class GenshiFormatter(Formatter):
    """Returns a Genshi event stream containing HTML formatting around the
    matched terms.
    """
    
    def __init__(self, qname="strong", between="..."):
        """
        :param qname: the QName for the tag to wrap around matched terms.
        :param between: the text to add between fragments.
        """
        
        self.qname = qname
        self.between = between
        
        from genshi.core import START, END, TEXT, Attrs, Stream
        self.START, self.END, self.TEXT = START, END, TEXT
        self.Attrs, self.Stream = Attrs, Stream

    def _add_text(self, text, output):
        if output and output[-1][0] == self.TEXT:
            output[-1] = (self.TEXT, output[-1][1] + text, output[-1][2])
        else:
            output.append((self.TEXT, text, (None, -1, -1)))

    def _format_fragment(self, text, fragment):
        START, TEXT, END, Attrs = self.START, self.TEXT, self.END, self.Attrs
        qname = self.qname
        output = []
        
        index = fragment.startchar
        lastmatched = False
        for t in fragment.matches:
            if t.startchar > index:
                if lastmatched:
                    output.append((END, qname, (None, -1, -1)))
                    lastmatched = False
                self._add_text(text[index:t.startchar], output)
            
            ttxt = text[t.startchar:t.endchar]
            if not lastmatched:
                output.append((START, (qname, Attrs()), (None, -1, -1)))
                lastmatched = True
            output.append((TEXT, ttxt, (None, -1, -1)))
                                    
            index = t.endchar
        
        if lastmatched:
            output.append((END, qname, (None, -1, -1)))
        
        return output

    def __call__(self, text, fragments):
        output = []
        first = True
        for fragment in fragments:
            if not first:
                self._add_text(self.between, output)
            first = False
            output += self._format_fragment(text, fragment)
        
        return self.Stream(output)


# Highlighting

def top_fragments(text, terms, analyzer, fragmenter, top=3,
                  scorer=None, minscore=1):
    if scorer is None:
        scorer = BasicFragmentScorer()
    
    termset = frozenset(terms)
    tokens = copyandmatchfilter(termset, analyzer(text, chars=True,
                                                  keeporiginal=True))
    scored_frags = nlargest(top, ((scorer(f), f)
                                  for f in fragmenter(text, tokens)))
    return [sf for score, sf in scored_frags if score > minscore]


def highlight(text, terms, analyzer, fragmenter, formatter, top=3,
              scorer=None, minscore=1, order=FIRST):
    
    if scorer is None:
        scorer = BasicFragmentScorer()
    
    if type(fragmenter) is type:
        fragmenter = fragmenter()
    if type(formatter) is type:
        formatter = formatter()
    if type(scorer) is type:
        scorer = scorer()
    
    fragments = top_fragments(text, terms, analyzer, fragmenter,
                              top=top, scorer=scorer, minscore=minscore)
    fragments.sort(key=order)
    return formatter(text, fragments)
    

if __name__ == '__main__':
    pass







from __future__ import with_statement
import inspect
from datetime import datetime

from nose.tools import assert_equal

from whoosh import analysis, fields, formats, qparser, query
from whoosh.filedb.filestore import RamStorage
from whoosh.qparser import dateparse
from whoosh.support.times import adatetime


def _plugin_classes(ignore):
    # Get all the subclasses of Weighting in whoosh.scoring
    return [c for name, c in inspect.getmembers(qparser, inspect.isclass)
            if qparser.Plugin in c.__bases__ and c not in ignore]


def test_combos():
    qs = 'w:a "hi there"^4.2 AND x:b^2.3 OR c AND (y:d OR e) (apple ANDNOT bear)^2.3'
    
    init_args = {qparser.DisMaxPlugin: ({"content": 1.0, "title": 1.2}, ),
                 qparser.FieldAliasPlugin: ({"content": ("text", "body")}, ),
                 qparser.MultifieldPlugin: (["title", "content"], ),
                 qparser.CopyFieldPlugin: ({"name": "phone"}, ),
                 }
    
    plugins = _plugin_classes(())
    for i, plugin in enumerate(plugins):
        try:
            plugins[i] = plugin(*init_args.get(plugin, ()))
        except TypeError:
            raise TypeError("Error instantiating %s" % plugin)
    
    count = 0
    for i, first in enumerate(plugins):
        for j in xrange(len(plugins)):
            if i == j: continue
            plist = [p for p in plugins[:j] if p is not first] + [first]
            qp = qparser.QueryParser("text", None, plugins=plist)
            try:
                qp.parse(qs)
            except Exception, e:
                raise Exception(str(e) + " combo: %s %r" % (count, plist))
            count += 1

def test_field_alias():
    qp = qparser.QueryParser("content", None)
    qp.add_plugin(qparser.FieldAliasPlugin({"title": ("article", "caption")}))
    q = qp.parse("alfa title:bravo article:charlie caption:delta")
    assert_equal(unicode(q), u"(content:alfa AND title:bravo AND title:charlie AND title:delta)")

def test_dateparser():
    schema = fields.Schema(text=fields.TEXT, date=fields.DATETIME)
    qp = qparser.QueryParser("text", schema)
    
    errs = []
    def cb(arg):
        errs.append(arg)
    basedate = datetime(2010, 9, 20, 15, 16, 6, 454000)
    qp.add_plugin(dateparse.DateParserPlugin(basedate, callback=cb))
    
    q = qp.parse(u"hello date:'last tuesday'")
    assert_equal(q.__class__, query.And)
    assert_equal(q[1].__class__, query.DateRange)
    assert_equal(q[1].startdate, adatetime(2010, 9, 14).floor())
    assert_equal(q[1].enddate, adatetime(2010, 9, 14).ceil())
    
    q = qp.parse(u"date:'3am to 5pm'")
    assert_equal(q.__class__, query.DateRange)
    assert_equal(q.startdate, adatetime(2010, 9, 20, 3).floor())
    assert_equal(q.enddate, adatetime(2010, 9, 20, 17).ceil())
    
    q = qp.parse(u"date:blah")
    assert_equal(q, query.NullQuery)
    assert_equal(errs[0], "blah")
    
    q = qp.parse(u"hello date:blarg")
    assert_equal(q.__class__, query.Term)
    assert_equal(q.fieldname, "text")
    assert_equal(q.text, "hello")
    assert_equal(errs[1], "blarg")
    
    q = qp.parse(u"hello date:20055x10")
    assert_equal(q.__class__, query.Term)
    assert_equal(q.fieldname, "text")
    assert_equal(q.text, "hello")
    assert_equal(errs[2], "20055x10")
    
    q = qp.parse(u"hello date:'2005 19 32'")
    assert_equal(q.__class__, query.Term)
    assert_equal(q.fieldname, "text")
    assert_equal(q.text, "hello")
    assert_equal(errs[3], "2005 19 32")
    
    q = qp.parse(u"date:'march 24 to dec 12'")
    assert_equal(q.__class__, query.DateRange)
    assert_equal(q.startdate, adatetime(2010, 3, 24).floor())
    assert_equal(q.enddate, adatetime(2010, 12, 12).ceil())
    
    q = qp.parse(u"date:('30 june' OR '10 july') quick")
    assert_equal(q.__class__, query.And)
    assert_equal(len(q), 2)
    assert_equal(q[0].__class__, query.Or)
    assert_equal(q[0][0].__class__, query.DateRange)
    assert_equal(q[0][1].__class__, query.DateRange)
    
def test_date_range():
    schema = fields.Schema(text=fields.TEXT, date=fields.DATETIME)
    qp = qparser.QueryParser("text", schema)
    basedate = datetime(2010, 9, 20, 15, 16, 6, 454000)
    qp.add_plugin(dateparse.DateParserPlugin(basedate))
    
    q = qp.parse(u"date:['30 march' to 'next wednesday']")
    assert_equal(q.__class__, query.DateRange)
    assert_equal(q.startdate, adatetime(2010, 3, 30).floor())
    assert_equal(q.enddate, adatetime(2010, 9, 22).ceil())
    
    q = qp.parse(u"date:[to 'next wednesday']")
    assert_equal(q.__class__, query.DateRange)
    assert_equal(q.startdate, None)
    assert_equal(q.enddate, adatetime(2010, 9, 22).ceil())
    
    q = qp.parse(u"date:['30 march' to]")
    assert_equal(q.__class__, query.DateRange)
    assert_equal(q.startdate, adatetime(2010, 3, 30).floor())
    assert_equal(q.enddate, None)
    
    q = qp.parse(u"date:[30 march to next wednesday]")
    assert_equal(q.__class__, query.DateRange)
    assert_equal(q.startdate, adatetime(2010, 3, 30).floor())
    assert_equal(q.enddate, adatetime(2010, 9, 22).ceil())
    
    q = qp.parse(u"date:[to next wednesday]")
    assert_equal(q.__class__, query.DateRange)
    assert_equal(q.startdate, None)
    assert_equal(q.enddate, adatetime(2010, 9, 22).ceil())
    
    q = qp.parse(u"date:[30 march to]")
    assert_equal(q.__class__, query.DateRange)
    assert_equal(q.startdate, adatetime(2010, 3, 30).floor())
    assert_equal(q.enddate, None)

def test_daterange_empty_field():
    schema = fields.Schema(test=fields.DATETIME)
    ix = RamStorage().create_index(schema)
                    
    writer = ix.writer()
    writer.add_document(test=None)
    writer.commit()
    
    with ix.searcher() as s:
        q = query.DateRange("test", datetime.fromtimestamp(0), datetime.today())
        r = s.search(q)
        assert_equal(len(r), 0)

def test_free_dates():
    a = analysis.StandardAnalyzer(stoplist=None)
    schema = fields.Schema(text=fields.TEXT(analyzer=a), date=fields.DATETIME)
    qp = qparser.QueryParser("text", schema)
    basedate = datetime(2010, 9, 20, 15, 16, 6, 454000)
    qp.add_plugin(dateparse.DateParserPlugin(basedate, free=True))
    
    q = qp.parse(u"hello date:last tuesday")
    assert_equal(q.__class__, query.And)
    assert_equal(len(q), 2)
    assert_equal(q[0].__class__, query.Term)
    assert_equal(q[0].text, "hello")
    assert_equal(q[1].__class__, query.DateRange)
    assert_equal(q[1].startdate, adatetime(2010, 9, 14).floor())
    assert_equal(q[1].enddate, adatetime(2010, 9, 14).ceil())
    
    q = qp.parse(u"date:mar 29 1972 hello")
    assert_equal(q.__class__, query.And)
    assert_equal(len(q), 2)
    assert_equal(q[0].__class__, query.DateRange)
    assert_equal(q[0].startdate, adatetime(1972, 3, 29).floor())
    assert_equal(q[0].enddate, adatetime(1972, 3, 29).ceil())
    assert_equal(q[1].__class__, query.Term)
    assert_equal(q[1].text, "hello")

    q = qp.parse(u"date:2005 march 2")
    assert_equal(q.__class__, query.DateRange)
    assert_equal(q.startdate, adatetime(2005, 3, 2).floor())
    assert_equal(q.enddate, adatetime(2005, 3, 2).ceil())
    
    q = qp.parse(u"date:'2005' march 2")
    assert_equal(q.__class__, query.And)
    assert_equal(len(q), 3)
    assert_equal(q[0].__class__, query.DateRange)
    assert_equal(q[0].startdate, adatetime(2005).floor())
    assert_equal(q[0].enddate, adatetime(2005).ceil())
    assert_equal(q[1].__class__, query.Term)
    assert_equal(q[1].fieldname, "text")
    assert_equal(q[1].text, "march")
    
    q = qp.parse(u"date:march 24 to dec 12")
    assert_equal(q.__class__, query.DateRange)
    assert_equal(q.startdate, adatetime(2010, 3, 24).floor())
    assert_equal(q.enddate, adatetime(2010, 12, 12).ceil())
    
    q = qp.parse(u"date:5:10pm")
    assert_equal(q.__class__, query.DateRange)
    assert_equal(q.startdate, adatetime(2010, 9, 20, 17, 10).floor())
    assert_equal(q.enddate, adatetime(2010, 9, 20, 17, 10).ceil())
    
    q = qp.parse(u"(date:30 june OR date:10 july) quick")
    assert_equal(q.__class__, query.And)
    assert_equal(len(q), 2)
    assert_equal(q[0].__class__, query.Or)
    assert_equal(q[0][0].__class__, query.DateRange)
    assert_equal(q[0][1].__class__, query.DateRange)
    
def test_prefix_plugin():
    schema = fields.Schema(id=fields.ID, text=fields.TEXT)
    ix = RamStorage().create_index(schema)
    
    w = ix.writer()
    w.add_document(id=u"1", text=u"alfa")
    w.add_document(id=u"2", text=u"bravo")
    w.add_document(id=u"3", text=u"buono")
    w.commit()
    
    with ix.searcher() as s:
        qp = qparser.QueryParser("text", schema)
        qp.remove_plugin_class(qparser.WildcardPlugin)
        qp.add_plugin(qparser.PrefixPlugin)
        
        q = qp.parse(u"b*")
        r = s.search(q, limit=None)
        assert_equal(len(r), 2)
        
        q = qp.parse(u"br*")
        r = s.search(q, limit=None)
        assert_equal(len(r), 1)
    
def test_custom_tokens():
    qp = qparser.QueryParser("text", None)
    qp.remove_plugin_class(qparser.CompoundsPlugin)
    qp.remove_plugin_class(qparser.NotPlugin)
    
    cp = qparser.CompoundsPlugin(And="&", Or="\\|", AndNot="&!", AndMaybe="&~", Not=None)
    qp.add_plugin(cp)
    
    np = qparser.NotPlugin("-")
    qp.add_plugin(np)
    
    q = qp.parse("this | that")
    assert_equal(q.__class__, query.Or)
    assert_equal(q[0].__class__, query.Term)
    assert_equal(q[0].text, "this")
    assert_equal(q[1].__class__, query.Term)
    assert_equal(q[1].text, "that")
    
    q = qp.parse("this&!that")
    assert_equal(q.__class__, query.AndNot)
    assert_equal(q.a.__class__, query.Term)
    assert_equal(q.a.text, "this")
    assert_equal(q.b.__class__, query.Term)
    assert_equal(q.b.text, "that")
    
    q = qp.parse("alfa -bravo NOT charlie")
    assert_equal(len(q), 4)
    assert_equal(q[1].__class__, query.Not)
    assert_equal(q[1].query.text, "bravo")
    assert_equal(q[2].text, "NOT")
    
def test_copyfield():
    qp = qparser.QueryParser("a", None)
    qp.add_plugin(qparser.CopyFieldPlugin({"b": "c"}))
    assert_equal(unicode(qp.parse("hello b:matt")), "(a:hello AND (b:matt OR c:matt))")
    
    qp = qparser.QueryParser("a", None)
    qp.add_plugin(qparser.CopyFieldPlugin({"b": "c"}, qparser.AndMaybeGroup))
    assert_equal(unicode(qp.parse("hello b:matt")), "(a:hello AND (b:matt ANDMAYBE c:matt))")
    
    qp = qparser.QueryParser("a", None)
    qp.add_plugin(qparser.CopyFieldPlugin({"b": "c"}, qparser.RequireGroup))
    assert_equal(unicode(qp.parse("hello (there OR b:matt)")), "(a:hello AND (a:there OR (b:matt REQUIRE c:matt)))")
    
    qp = qparser.QueryParser("a", None)
    qp.add_plugin(qparser.CopyFieldPlugin({"a": "c"}))
    assert_equal(unicode(qp.parse("hello there")), "((a:hello OR c:hello) AND (a:there OR c:there))")
    
    qp = qparser.QueryParser("a", None)
    qp.add_plugin(qparser.CopyFieldPlugin({"b": "c"}, mirror=True))
    assert_equal(unicode(qp.parse("hello c:matt")), "(a:hello AND (c:matt OR b:matt))")
    
    qp = qparser.QueryParser("a", None)
    qp.add_plugin(qparser.CopyFieldPlugin({"c": "a"}, mirror=True))
    assert_equal(unicode(qp.parse("hello c:matt")), "((a:hello OR c:hello) AND (c:matt OR a:matt))")
    
    ana = analysis.RegexAnalyzer(r"\w+") | analysis.DoubleMetaphoneFilter()
    fmt = formats.Frequency(ana)
    schema = fields.Schema(name=fields.KEYWORD, name_phone=fields.FieldType(fmt, multitoken_query="or"))
    qp = qparser.QueryParser("name", schema)
    qp.add_plugin(qparser.CopyFieldPlugin({"name": "name_phone"}))
    assert_equal(unicode(qp.parse(u"spruce view")), "((name:spruce OR name_phone:SPRS) AND (name:view OR name_phone:F OR name_phone:FF))")
    
def test_gtlt():
    schema = fields.Schema(a=fields.KEYWORD, b=fields.NUMERIC,
                           c=fields.KEYWORD,
                           d=fields.NUMERIC(float), e=fields.DATETIME)
    qp = qparser.QueryParser("a", schema)
    qp.add_plugin(qparser.GtLtPlugin())
    qp.add_plugin(qparser.dateparse.DateParserPlugin())
    
    q = qp.parse(u"a:hello b:>100 c:<=z there")
    assert_equal(q.__class__, query.And)
    assert_equal(len(q), 4)
    assert_equal(q[0], query.Term("a", "hello"))
    assert_equal(q[1], query.NumericRange("b", 100, None, startexcl=True))
    assert_equal(q[2], query.TermRange("c", None, 'z'))
    assert_equal(q[3], query.Term("a", "there"))
    
    q = qp.parse(u"hello e:>'29 mar 2001' there")
    assert_equal(q.__class__, query.And)
    assert_equal(len(q), 3)
    assert_equal(q[0], query.Term("a", "hello"))
    # As of this writing, date ranges don't support startexcl/endexcl
    assert_equal(q[1], query.DateRange("e", datetime(2001, 3, 29, 0, 0), None))
    assert_equal(q[2], query.Term("a", "there"))
    
    q = qp.parse(u"a:> alfa c:<= bravo")
    assert_equal(unicode(q), "(a:a: AND a:alfa AND a:c: AND a:bravo)")
    
    qp.remove_plugin_class(qparser.FieldsPlugin)
    qp.remove_plugin_class(qparser.RangePlugin)
    q = qp.parse(u"hello a:>500 there")
    assert_equal(unicode(q), "(a:hello AND a:a: AND a:500 AND a:there)")
        
        










from __future__ import with_statement
from nose.tools import assert_equal, assert_raises
import random
from collections import defaultdict

from whoosh import fields, query
from whoosh.filedb.filestore import RamStorage
from whoosh.filedb.filewriting import NO_MERGE
from whoosh.util import length_to_byte, byte_to_length, permutations
from whoosh.writing import IndexingError
from whoosh.support.testing import skip_if_unavailable, TempIndex, skip_if

def no_queue_support():
    try:
        import multiprocessing.synchronize
    except ImportError:
        return True
    else:
        try:
            from multiprocessing import Queue
            Queue()
        except OSError:
            return True
        else:
            return False


def test_creation():
    s = fields.Schema(content=fields.TEXT(phrase=True),
                      title=fields.TEXT(stored=True),
                      path=fields.ID(stored=True),
                      tags=fields.KEYWORD(stored=True),
                      quick=fields.NGRAM,
                      note=fields.STORED)
    st = RamStorage()
    
    ix = st.create_index(s)
    w = ix.writer()
    w.add_document(title=u"First", content=u"This is the first document", path=u"/a",
                   tags=u"first second third", quick=u"First document", note=u"This is the first document")
    w.add_document(content=u"Let's try this again", title=u"Second", path=u"/b",
                   tags=u"Uno Dos Tres", quick=u"Second document", note=u"This is the second document")
    w.commit()

def test_empty_commit():
    s = fields.Schema(id=fields.ID(stored=True))
    with TempIndex(s, "emptycommit") as ix:
        w = ix.writer()
        w.add_document(id=u"1")
        w.add_document(id=u"2")
        w.add_document(id=u"3")
        w.commit()
        
        w = ix.writer()
        w.commit()

def _check_writer(name, writer_fn):
    schema = fields.Schema(text=fields.TEXT, id=fields.STORED)
    domain = (u"alfa", u"bravo", u"charlie", u"delta", u"echo", u"foxtrot", u"golf",
              u"hotel", u"india", u"juliet", u"kilo", u"lima", u"mike", u"november")
    docs = defaultdict(list)
    with TempIndex(schema, name) as ix:
        w = writer_fn(ix)
        for i in xrange(1000):
            smp = random.sample(domain, 5)
            for word in smp:
                docs[word].append(i)
            w.add_document(text=u" ".join(smp), id=i)
        w.commit()
        
        with ix.searcher() as s:
            for word in domain:
                print word
                rset = sorted([hit["id"] for hit in s.search(query.Term("text", word), limit=None)])
                assert_equal(rset, docs[word])

def test_simple():
    _check_writer("simplew", lambda ix: ix.writer())

@skip_if_unavailable("multiprocessing")
@skip_if(no_queue_support)
def test_multipool():
    _check_writer("multipool", lambda ix: ix.writer(procs=4))

@skip_if_unavailable("multiprocessing")
@skip_if(no_queue_support)
def test_multisegwriter():
    from whoosh.filedb.multiproc import MultiSegmentWriter
    _check_writer("multisegw", lambda ix: MultiSegmentWriter(ix, procs=4))

def test_integrity():
    s = fields.Schema(name=fields.TEXT, value=fields.TEXT)
    st = RamStorage()
    ix = st.create_index(s)
    
    w = ix.writer()
    w.add_document(name=u"Yellow brown", value=u"Blue red green purple?")
    w.add_document(name=u"Alpha beta", value=u"Gamma delta epsilon omega.")
    w.commit()
    
    w = ix.writer()
    w.add_document(name=u"One two", value=u"Three four five.")
    w.commit()
    
    tr = ix.reader()
    assert_equal(ix.doc_count_all(), 3)
    assert_equal(list(tr.lexicon("name")), ["alpha", "beta", "brown", "one", "two", "yellow"])

def test_lengths():
    s = fields.Schema(f1=fields.KEYWORD(stored=True, scorable=True),
                      f2=fields.KEYWORD(stored=True, scorable=True))
    with TempIndex(s, "testlengths") as ix:
        w = ix.writer()
        tokens = u"ABCDEFG"
        from itertools import cycle, islice
        lengths = [10, 20, 2, 102, 45, 3, 420, 2]
        for length in lengths:
            w.add_document(f2=u" ".join(islice(cycle(tokens), length)))
        w.commit()
        
        with ix.reader() as dr:
            ls1 = [dr.doc_field_length(i, "f1") for i in xrange(0, len(lengths))]
            assert_equal(ls1, [0] * len(lengths))
            ls2 = [dr.doc_field_length(i, "f2") for i in xrange(0, len(lengths))]
            assert_equal(ls2, [byte_to_length(length_to_byte(l))for l in lengths])

def test_lengths_ram():
    s = fields.Schema(f1=fields.KEYWORD(stored=True, scorable=True),
                      f2=fields.KEYWORD(stored=True, scorable=True))
    st = RamStorage()
    ix = st.create_index(s)
    w = ix.writer()
    w.add_document(f1=u"A B C D E", f2=u"X Y Z")
    w.add_document(f1=u"B B B B C D D Q", f2=u"Q R S T")
    w.add_document(f1=u"D E F", f2=u"U V A B C D E")
    w.commit()
    
    dr = ix.reader()
    assert_equal(dr.stored_fields(0)["f1"], "A B C D E")
    assert_equal(dr.doc_field_length(0, "f1"), 5)
    assert_equal(dr.doc_field_length(1, "f1"), 8)
    assert_equal(dr.doc_field_length(2, "f1"), 3)
    assert_equal(dr.doc_field_length(0, "f2"), 3)
    assert_equal(dr.doc_field_length(1, "f2"), 4)
    assert_equal(dr.doc_field_length(2, "f2"), 7)
    
    assert_equal(dr.field_length("f1"), 16)
    assert_equal(dr.field_length("f2"), 14)
    assert_equal(dr.max_field_length("f1"), 8)
    assert_equal(dr.max_field_length("f2"), 7)
    
def test_merged_lengths():
    s = fields.Schema(f1=fields.KEYWORD(stored=True, scorable=True),
                      f2=fields.KEYWORD(stored=True, scorable=True))
    with TempIndex(s, "mergedlengths") as ix:
        w = ix.writer()
        w.add_document(f1=u"A B C", f2=u"X")
        w.add_document(f1=u"B C D E", f2=u"Y Z")
        w.commit()
        
        w = ix.writer()
        w.add_document(f1=u"A", f2=u"B C D E X Y")
        w.add_document(f1=u"B C", f2=u"X")
        w.commit(NO_MERGE)
        
        w = ix.writer()
        w.add_document(f1=u"A B X Y Z", f2=u"B C")
        w.add_document(f1=u"Y X", f2=u"A B")
        w.commit(NO_MERGE)
        
        with ix.reader() as dr:
            assert_equal(dr.stored_fields(0)["f1"], u"A B C")
            assert_equal(dr.doc_field_length(0, "f1"), 3)
            assert_equal(dr.doc_field_length(2, "f2"), 6)
            assert_equal(dr.doc_field_length(4, "f1"), 5)
    
def test_frequency_keyword():
    s = fields.Schema(content=fields.KEYWORD)
    st = RamStorage()
    ix = st.create_index(s)
    
    w = ix.writer()
    w.add_document(content=u"A B C D E")
    w.add_document(content=u"B B B B C D D")
    w.add_document(content=u"D E F")
    w.commit()
    
    tr = ix.reader()
    assert_equal(tr.doc_frequency("content", u"B"), 2)
    assert_equal(tr.frequency("content", u"B"), 5)
    assert_equal(tr.doc_frequency("content", u"E"), 2)
    assert_equal(tr.frequency("content", u"E"), 2)
    assert_equal(tr.doc_frequency("content", u"A"), 1)
    assert_equal(tr.frequency("content", u"A"), 1)
    assert_equal(tr.doc_frequency("content", u"D"), 3)
    assert_equal(tr.frequency("content", u"D"), 4)
    assert_equal(tr.doc_frequency("content", u"F"), 1)
    assert_equal(tr.frequency("content", u"F"), 1)
    assert_equal(tr.doc_frequency("content", u"Z"), 0)
    assert_equal(tr.frequency("content", u"Z"), 0)
    assert_equal(list(tr), [("content", u"A", 1, 1), ("content", u"B", 2, 5),
                            ("content", u"C", 2, 2), ("content", u"D", 3, 4),
                            ("content", u"E", 2, 2), ("content", u"F", 1, 1)])
    tr.close()
    
def test_frequency_text():
    s = fields.Schema(content=fields.KEYWORD)
    st = RamStorage()
    ix = st.create_index(s)
    
    w = ix.writer()
    w.add_document(content=u"alfa bravo charlie delta echo")
    w.add_document(content=u"bravo bravo bravo bravo charlie delta delta")
    w.add_document(content=u"delta echo foxtrot")
    w.commit()
    
    with ix.reader() as tr:
        assert_equal(tr.doc_frequency("content", u"bravo"), 2)
        assert_equal(tr.frequency("content", u"bravo"), 5)
        assert_equal(tr.doc_frequency("content", u"echo"), 2)
        assert_equal(tr.frequency("content", u"echo"), 2)
        assert_equal(tr.doc_frequency("content", u"alfa"), 1)
        assert_equal(tr.frequency("content", u"alfa"), 1)
        assert_equal(tr.doc_frequency("content", u"delta"), 3)
        assert_equal(tr.frequency("content", u"delta"), 4)
        assert_equal(tr.doc_frequency("content", u"foxtrot"), 1)
        assert_equal(tr.frequency("content", u"foxtrot"), 1)
        assert_equal(tr.doc_frequency("content", u"zulu"), 0)
        assert_equal(tr.frequency("content", u"zulu"), 0)
        assert_equal(list(tr), [("content", u"alfa", 1, 1), ("content", u"bravo", 2, 5),
                                ("content", u"charlie", 2, 2), ("content", u"delta", 3, 4),
                                ("content", u"echo", 2, 2), ("content", u"foxtrot", 1, 1)])

def test_deletion():
    s = fields.Schema(key=fields.ID, name=fields.TEXT, value=fields.TEXT)
    with TempIndex(s, "deletion") as ix:
        w = ix.writer()
        w.add_document(key=u"A", name=u"Yellow brown", value=u"Blue red green purple?")
        w.add_document(key=u"B", name=u"Alpha beta", value=u"Gamma delta epsilon omega.")
        w.add_document(key=u"C", name=u"One two", value=u"Three four five.")
        w.commit()
        
        w = ix.writer()
        count = w.delete_by_term("key", u"B")
        assert_equal(count, 1)
        w.commit(merge=False)
        
        assert_equal(ix.doc_count_all(), 3)
        assert_equal(ix.doc_count(), 2)
        
        w = ix.writer()
        w.add_document(key=u"A", name=u"Yellow brown", value=u"Blue red green purple?")
        w.add_document(key=u"B", name=u"Alpha beta", value=u"Gamma delta epsilon omega.")
        w.add_document(key=u"C", name=u"One two", value=u"Three four five.")
        w.commit()
        
        # This will match both documents with key == B, one of which is already
        # deleted. This should not raise an error.
        w = ix.writer()
        count = w.delete_by_term("key", u"B")
        assert_equal(count, 1)
        w.commit()
        
        ix.optimize()
        assert_equal(ix.doc_count_all(), 4)
        assert_equal(ix.doc_count(), 4)
        
        with ix.reader() as tr:
            assert_equal(list(tr.lexicon("name")), ["brown", "one", "two", "yellow"])

def test_writer_reuse():
    s = fields.Schema(key=fields.ID)
    ix = RamStorage().create_index(s)
    
    w = ix.writer()
    w.add_document(key=u"A")
    w.add_document(key=u"B")
    w.add_document(key=u"C")
    w.commit()
    
    # You can't re-use a commited/canceled writer
    assert_raises(IndexingError, w.add_document, key=u"D")
    assert_raises(IndexingError, w.update_document, key=u"B")
    assert_raises(IndexingError, w.delete_document, 0)
    assert_raises(IndexingError, w.add_reader, None)
    assert_raises(IndexingError, w.add_field, "name", fields.ID)
    assert_raises(IndexingError, w.remove_field, "key")
    assert_raises(IndexingError, w.searcher)

def test_update():
    # Test update with multiple unique keys
    SAMPLE_DOCS = [{"id": u"test1", "path": u"/test/1", "text": u"Hello"},
                   {"id": u"test2", "path": u"/test/2", "text": u"There"},
                   {"id": u"test3", "path": u"/test/3", "text": u"Reader"},
                   ]
    
    schema = fields.Schema(id=fields.ID(unique=True, stored=True),
                           path=fields.ID(unique=True, stored=True),
                           text=fields.TEXT)
    
    with TempIndex(schema, "update") as ix:
        writer = ix.writer()
        for doc in SAMPLE_DOCS:
            writer.add_document(**doc)
        writer.commit()
        
        writer = ix.writer()
        writer.update_document(id=u"test2", path=u"test/1", text=u"Replacement")
        writer.commit()
    
def test_update2():
    schema = fields.Schema(key=fields.ID(unique=True, stored=True),
                           p=fields.ID(stored=True))
    with TempIndex(schema, "update2") as ix:
        nums = range(100)
        random.shuffle(nums)
        for i, n in enumerate(nums):
            w = ix.writer()
            w.update_document(key=unicode(n % 10), p=unicode(i))
            w.commit()
            
        with ix.searcher() as s:
            results = [d["key"] for d in s.all_stored_fields()]
            results.sort()
            assert_equal(results, ["0", "1", "2", "3", "4", "5", "6", "7", "8", "9"])

def test_reindex():
    SAMPLE_DOCS = [
        {'id': u'test1', 'text': u'This is a document. Awesome, is it not?'},
        {'id': u'test2', 'text': u'Another document. Astounding!'},
        {'id': u'test3', 'text': u'A fascinating article on the behavior of domestic steak knives.'},
    ]

    schema = fields.Schema(text=fields.TEXT(stored=True),
                           id=fields.ID(unique=True, stored=True))
    with TempIndex(schema, "reindex") as ix:
        def reindex():
            writer = ix.writer()
            for doc in SAMPLE_DOCS:
                writer.update_document(**doc)
            writer.commit()

        reindex()
        assert_equal(ix.doc_count_all(), 3)
        reindex()
        assert_equal(ix.doc_count_all(), 3)
    
def test_noscorables1():
    values = [u"alfa", u"bravo", u"charlie", u"delta", u"echo", u"foxtrot",
              u"golf", u"hotel", u"india", u"juliet", u"kilo", u"lima"]
    from random import choice, sample, randint
    
    times = 1000
    
    schema = fields.Schema(id=fields.ID, tags=fields.KEYWORD)
    with TempIndex(schema, "noscorables1") as ix:
        w = ix.writer()
        for _ in xrange(times):
            w.add_document(id=choice(values), tags=u" ".join(sample(values, randint(2, 7))))
        w.commit()
        
        with ix.searcher() as s:
            s.search(query.Term("id", "bravo"))
        
def test_noscorables2():
    schema = fields.Schema(field=fields.ID)
    with TempIndex(schema, "noscorables2") as ix:
        writer = ix.writer()
        writer.add_document(field=u'foo')
        writer.commit()

def test_multi():
    schema = fields.Schema(id=fields.ID(stored=True),
                           content=fields.KEYWORD(stored=True))
    with TempIndex(schema, "multi") as ix:
        writer = ix.writer()
        writer.add_document(id=u"1", content=u"alfa bravo charlie") #deleted 1
        writer.add_document(id=u"2", content=u"bravo charlie delta echo") #deleted 1
        writer.add_document(id=u"3", content=u"charlie delta echo foxtrot") #deleted 2
        writer.commit()
        
        writer = ix.writer()
        writer.delete_by_term("id", "1")
        writer.delete_by_term("id", "2")
        writer.add_document(id=u"4", content=u"apple bear cherry donut")
        writer.add_document(id=u"5", content=u"bear cherry donut eggs")
        writer.add_document(id=u"6", content=u"delta echo foxtrot golf") #deleted 2
        writer.add_document(id=u"7", content=u"echo foxtrot golf hotel") # no d
        writer.commit(merge=False)
        
        writer = ix.writer()
        writer.delete_by_term("id", "3")
        writer.delete_by_term("id", "6")
        writer.add_document(id=u"8", content=u"cherry donut eggs falafel")
        writer.add_document(id=u"9", content=u"donut eggs falafel grape")
        writer.add_document(id=u"A", content=u" foxtrot golf hotel india")
        writer.commit(merge=False)

        assert_equal(ix.doc_count(), 6)
    
        with ix.searcher() as s:
            r = s.search(query.Prefix("content", u"d"), optimize=False)
            assert_equal(sorted([d["id"] for d in r]), ["4", "5", "8", "9"])
            
            r = s.search(query.Prefix("content", u"d"))
            assert_equal(sorted([d["id"] for d in r]), ["4", "5", "8", "9"])
            
            r = s.search(query.Prefix("content", u"d"), limit=None)
            assert_equal(sorted([d["id"] for d in r]), ["4", "5", "8", "9"])
    
def test_deleteall():
    schema = fields.Schema(text=fields.TEXT)
    with TempIndex(schema, "deleteall") as ix:
        w = ix.writer()
        domain = u"alfa bravo charlie delta echo".split()
        for i, ls in enumerate(permutations(domain)):
            w.add_document(text=u" ".join(ls))
            if not i % 10:
                w.commit()
                w = ix.writer()
        w.commit()
        
        # This is just a test, don't use this method to delete all docs IRL!
        doccount = ix.doc_count_all()
        w = ix.writer()
        for docnum in xrange(doccount):
            w.delete_document(docnum)
        w.commit()
        
        with ix.searcher() as s:
            r = s.search(query.Or([query.Term("text", u"alfa"), query.Term("text", u"bravo")]))
            assert_equal(len(r), 0)
        
        ix.optimize()
        assert_equal(ix.doc_count_all(), 0)
        
        with ix.reader() as r:
            assert_equal(list(r), [])
            
def test_single():
    schema = fields.Schema(id=fields.ID(stored=True), text=fields.TEXT)
    with TempIndex(schema, "single") as ix:
        w = ix.writer()
        w.add_document(id=u"1", text=u"alfa")
        w.commit()
        
        with ix.searcher() as s:
            assert ("text", u"alfa") in s.reader()
            assert_equal(list(s.documents(id="1")), [{"id": "1"}])
            assert_equal(list(s.documents(text="alfa")), [{"id": "1"}])
            assert_equal(list(s.all_stored_fields()), [{"id": "1"}])
    
def test_indentical_fields():
    schema = fields.Schema(id=fields.STORED,
                           f1=fields.TEXT, f2=fields.TEXT, f3=fields.TEXT)
    with TempIndex(schema, "identifields") as ix:
        w = ix.writer()
        w.add_document(id=1, f1=u"alfa", f2=u"alfa", f3 = u"alfa")
        w.commit()
    
        with ix.searcher() as s:
            assert_equal(list(s.lexicon("f1")), ["alfa"])
            assert_equal(list(s.lexicon("f2")), ["alfa"])
            assert_equal(list(s.lexicon("f3")), ["alfa"])
            assert_equal(list(s.documents(f1="alfa")), [{"id": 1}])
            assert_equal(list(s.documents(f2="alfa")), [{"id": 1}])
            assert_equal(list(s.documents(f3="alfa")), [{"id": 1}])
        
            
                





from nose.tools import assert_equal

from whoosh import fields, qparser
from whoosh.analysis import *
from whoosh.filedb.filestore import RamStorage


def test_regextokenizer():
    value = u"AAAaaaBBBbbbCCCcccDDDddd"
    
    rex = RegexTokenizer("[A-Z]+")
    assert_equal([t.text for t in rex(value)], [u"AAA", u"BBB", u"CCC", u"DDD"])
    
    rex = RegexTokenizer("[A-Z]+", gaps=True)
    assert_equal([t.text for t in rex(value)], [u"aaa", u"bbb", u"ccc", u"ddd"])

def test_composition1():
    ca = RegexTokenizer() | LowercaseFilter()
    assert_equal(ca.__class__.__name__, "CompositeAnalyzer")
    assert_equal(ca[0].__class__.__name__, "RegexTokenizer")
    assert_equal(ca[1].__class__.__name__, "LowercaseFilter")
    assert_equal([t.text for t in ca(u"ABC 123")], ["abc", "123"])

def test_composition2():
    ca = RegexTokenizer() | LowercaseFilter()
    sa = ca | StopFilter()
    assert_equal(len(sa), 3)
    assert_equal(sa.__class__.__name__, "CompositeAnalyzer")
    assert_equal(sa[0].__class__.__name__, "RegexTokenizer")
    assert_equal(sa[1].__class__.__name__, "LowercaseFilter")
    assert_equal(sa[2].__class__.__name__, "StopFilter")
    assert_equal([t.text for t in sa(u"The ABC 123")], ["abc", "123"])

def test_composition3():
    sa = RegexTokenizer() | StopFilter()
    assert_equal(sa.__class__.__name__, "CompositeAnalyzer")

def test_composing_functions():
    def filter(tokens):
        for t in tokens:
            t.text = t.text.upper()
            yield t
            
    analyzer = RegexTokenizer() | filter
    assert_equal([t.text for t in analyzer(u"abc def")], ["ABC", "DEF"])

def test_shared_composition():
    shared = RegexTokenizer(r"\S+") | LowercaseFilter()
    
    ana1 = shared | NgramFilter(3)
    ana2 = shared | DoubleMetaphoneFilter()
    
    assert_equal([t.text for t in ana1(u"hello")], ["hel", "ell", "llo"])
    assert_equal([t.text for t in ana2(u"hello")], ["HL"])

def test_multifilter():
    f1 = LowercaseFilter()
    f2 = PassFilter()
    mf = MultiFilter(a=f1, b=f2)
    ana = RegexTokenizer(r"\S+") | mf
    text = u"ALFA BRAVO CHARLIE"
    assert_equal([t.text for t in ana(text, mode="a")], ["alfa", "bravo", "charlie"])
    assert_equal([t.text for t in ana(text, mode="b")], ["ALFA", "BRAVO", "CHARLIE"])

def test_intraword():
    iwf = IntraWordFilter(mergewords=True, mergenums=True)
    ana = RegexTokenizer(r"\S+") | iwf
    
    def check(text, ls):
        assert_equal([(t.pos, t.text) for t in ana(text)], ls)
        
    check(u"PowerShot", [(0, "Power"), (1, "Shot"), (1, "PowerShot")])
    check(u"A's+B's&C's", [(0, "A"), (1, "B"), (2, "C"), (2, "ABC")])
    check(u"Super-Duper-XL500-42-AutoCoder!",
          [(0, "Super"), (1, "Duper"), (2, "XL"), (2, "SuperDuperXL"),
           (3, "500"), (4, "42"), (4, "50042"), (5, "Auto"), (6, "Coder"),
           (6, "AutoCoder")])

def test_biword():
    ana = RegexTokenizer(r"\w+") | BiWordFilter()
    result = [t.copy() for t
              in ana(u"the sign of four", chars=True, positions=True)]
    assert_equal(["the-sign", "sign-of", "of-four"], [t.text for t in result])
    assert_equal([(0, 8), (4, 11), (9, 16)], [(t.startchar, t.endchar) for t in result])
    assert_equal([0, 1, 2], [t.pos for t in result])
    
    result = [t.copy() for t in ana(u"single")]
    assert_equal(len(result), 1)
    assert_equal(result[0].text, "single")

def test_shingles():
    ana = RegexTokenizer(r"\w+") | ShingleFilter(3, " ")
    source = u"better a witty fool than a foolish wit"
    results = [t.copy() for t in ana(source, positions=True, chars=True)]
    assert_equal([t.text for t in results],
                 [u'better a witty', u'a witty fool', u'witty fool than',
                  u'fool than a', u'than a foolish', u'a foolish wit'])
    assert_equal([t.pos for t in results], range(len(results)))
    for t in results:
        assert_equal(t.text, source[t.startchar:t.endchar])
    
def test_unicode_blocks():
    from whoosh.support.unicode import blocks, blockname, blocknum
    
    assert_equal(blockname(u'a'), 'Basic Latin')
    assert_equal(blockname(unichr(0x0b80)), 'Tamil')
    assert_equal(blockname(unichr(2048)), None)
    assert_equal(blocknum(u'a'), 0)
    assert_equal(blocknum(unichr(0x0b80)), 22)
    assert_equal(blocknum(unichr(2048)), None)
    assert_equal(blocknum(u'a'), blocks.Basic_Latin)
    assert_equal(blocknum(unichr(0x0b80)), blocks.Tamil)
    
def test_double_metaphone():
    mf = RegexTokenizer() | LowercaseFilter() | DoubleMetaphoneFilter()
    results = [(t.text, t.boost) for t in mf(u"Spruce View")]
    assert_equal(results, [('SPRS', 1.0), ('F', 1.0), ('FF', 0.5)])
    
    mf = RegexTokenizer() | LowercaseFilter() | DoubleMetaphoneFilter(combine=True)
    results = [(t.text, t.boost) for t in mf(u"Spruce View")]
    assert_equal(results, [('spruce', 1.0), ('SPRS', 1.0), ('view', 1.0),
                           ('F', 1.0), ('FF', 0.5)])

    namefield = fields.TEXT(analyzer=mf)
    texts = list(namefield.process_text(u"Spruce View", mode="query"))
    assert_equal(texts, [u'spruce', 'SPRS', u'view', 'F', 'FF'])

def test_substitution():
    mf = RegexTokenizer(r"\S+") | SubstitutionFilter("-", "")
    assert_equal([t.text for t in mf(u"one-two th-re-ee four")],
                 ["onetwo", "threee", "four"])
    
    mf = RegexTokenizer(r"\S+") | SubstitutionFilter("([^=]*)=(.*)", r"\2=\1")
    assert_equal([t.text for t in mf(u"a=b c=d ef")], ["b=a", "d=c", "ef"])

def test_delimited_attribute():
    ana = RegexTokenizer(r"\S+") | DelimitedAttributeFilter()
    results = [(t.text, t.boost) for t in ana(u"image render^2 file^0.5")]
    assert_equal(results, [("image", 1.0), ("render", 2.0), ("file", 0.5)])
    
def test_porter2():
    from whoosh.lang.porter2 import stem
    
    plurals = ['caresses', 'flies', 'dies', 'mules', 'denied',
               'died', 'agreed', 'owned', 'humbled', 'sized',
               'meeting', 'stating', 'siezing', 'itemization',
               'sensational', 'traditional', 'reference', 'colonizer',
               'plotted']
    singles = [stem(w) for w in plurals]
    
    assert_equal(singles, ['caress', 'fli', 'die', 'mule', 'deni', 'die', 'agre',
                           'own', 'humbl', 'size', 'meet', 'state', 'siez', 'item',
                           'sensat', 'tradit', 'refer', 'colon', 'plot'])
    assert_equal(stem("bill's"), "bill")
    assert_equal(stem("y's"), "y")

def test_url():
    sample = u"Visit http://bitbucket.org/mchaput/whoosh or urn:isbn:5930502 or http://www.apple.com/."
    
    for ana in (SimpleAnalyzer(url_pattern),
                StandardAnalyzer(url_pattern, stoplist=None)):
        ts = [t.text for t in ana(sample)]
        assert_equal(ts, [u'visit', u'http://bitbucket.org/mchaput/whoosh',
                          u'or', u'urn:isbn:5930502', u'or', u'http://www.apple.com/'])

def test_name_field():
    ana = (RegexTokenizer(r"\S+")
           | LowercaseFilter()
           | DoubleMetaphoneFilter(combine=True))
    namefield = fields.TEXT(analyzer=ana, multitoken_query="or")
    schema = fields.Schema(id=fields.STORED, name=namefield)
    
    ix = RamStorage().create_index(schema)
    w = ix.writer()
    w.add_document(id=u"one", name=u"Leif Ericson")
    w.commit()
    
    s = ix.searcher()
    qp = qparser.QueryParser("name", schema)
    q = qp.parse(u"leaf eriksen", normalize=False)
    r = s.search(q)
    assert_equal(len(r), 1)








# encoding: utf8

from __future__ import with_statement
import random

from nose.tools import assert_equal

from whoosh.filedb.filestore import RamStorage
from whoosh.filedb.filetables import (HashReader, HashWriter,
                                      OrderedHashWriter, OrderedHashReader,
                                      StoredFieldWriter, StoredFieldReader,
                                      TermIndexWriter, TermIndexReader)
from whoosh.support.testing import TempStorage


def randstring(domain, minlen, maxlen):
    return "".join(random.sample(domain, random.randint(minlen, maxlen)))


def test_termkey():
    with TempStorage("termkey") as st:
        tw = TermIndexWriter(st.create_file("test.trm"))
        tw.add(("alfa", u"bravo"), (1.0, 2, 3))
        tw.add((u"alfa", u"Ã¦Ã¯Å�Ãº"), (4.0, 5, 6))
        tw.add((u"text", u"æ—¥æœ¬èªž"), (7.0, 8, 9))
        tw.close()
        
        tr = TermIndexReader(st.open_file("test.trm"))
        assert ("alfa", u"bravo") in tr
        assert (u"alfa", u"Ã¦Ã¯Å�Ãº") in tr
        assert (u"text", u"æ—¥æœ¬èªž") in tr
        tr.close()
    
def test_random_termkeys():
    def random_fieldname():
        return "".join(chr(random.randint(65, 90)) for _ in xrange(1, 20))
    
    def random_token():
        return "".join(unichr(random.randint(0, 0xd7ff)) for _ in xrange(1, 20))
    
    domain = sorted([(random_fieldname(), random_token()) for _ in xrange(1000)])
    
    st = RamStorage()
    tw = TermIndexWriter(st.create_file("test.trm"))
    for term in domain:
        tw.add(term, (1.0, 0, 1))
    tw.close()
    
    tr = TermIndexReader(st.open_file("test.trm"))
    for term in domain:
        assert term in tr
    
def test_hash():
    with TempStorage("hash") as st:
        hwf = st.create_file("test.hsh")
        hw = HashWriter(hwf)
        hw.add("foo", "bar")
        hw.add("glonk", "baz")
        hw.close()
        
        hrf = st.open_file("test.hsh")
        hr = HashReader(hrf)
        assert_equal(hr.get("foo"), "bar")
        assert_equal(hr.get("baz"), None)
        hr.close()

def test_hash_contents():
    samp = set((('alfa', 'bravo'), ('charlie', 'delta'), ('echo', 'foxtrot'),
               ('golf', 'hotel'), ('india', 'juliet'), ('kilo', 'lima'),
               ('mike', 'november'), ('oskar', 'papa'), ('quebec', 'romeo'),
               ('sierra', 'tango'), ('ultra', 'victor'), ('whiskey', 'xray')))
    
    with TempStorage("hashcontents") as st:
        hwf = st.create_file("test.hsh")
        hw = HashWriter(hwf)
        hw.add_all(samp)
        hw.close()
        
        hrf = st.open_file("test.hsh")
        hr = HashReader(hrf)
        assert_equal(set(hr.items()), samp)
        hr.close()
    
def test_random_hash():
    with TempStorage("randomhash") as st:
        domain = "abcdefghijklmnopqrstuvwxyz"
        domain += domain.upper()
        times = 1000
        minlen = 1
        maxlen = len(domain)
        
        samp = dict((randstring(domain, minlen, maxlen),
                     randstring(domain, minlen, maxlen)) for _ in xrange(times))
        
        hwf = st.create_file("test.hsh")
        hw = HashWriter(hwf)
        for k, v in samp.iteritems():
            hw.add(k, v)
        hw.close()
        
        keys = samp.keys()
        random.shuffle(keys)
        hrf = st.open_file("test.hsh")
        hr = HashReader(hrf)
        for k in keys:
            v = hr[k]
            assert_equal(v, samp[k])
        hr.close()

def test_ordered_hash():
    times = 10000
    with TempStorage("orderedhash") as st:
        hwf = st.create_file("test.hsh")
        hw = HashWriter(hwf)
        hw.add_all(("%08x" % x, str(x)) for x in xrange(times))
        hw.close()
        
        keys = range(times)
        random.shuffle(keys)
        hrf = st.open_file("test.hsh")
        hr = HashReader(hrf)
        for x in keys:
            assert_equal(hr["%08x" % x], str(x))
        hr.close()
    
def test_ordered_closest():
    keys = ['alfa', 'bravo', 'charlie', 'delta', 'echo', 'foxtrot', 'golf',
            'hotel', 'india', 'juliet', 'kilo', 'lima', 'mike', 'november']
    values = [''] * len(keys)
    
    with TempStorage("orderedclosest") as st:
        hwf = st.create_file("test.hsh")
        hw = OrderedHashWriter(hwf)
        hw.add_all(zip(keys, values))
        hw.close()
        
        hrf = st.open_file("test.hsh")
        hr = OrderedHashReader(hrf)
        ck = hr.closest_key
        assert_equal(ck(''), 'alfa')
        assert_equal(ck(' '), 'alfa')
        assert_equal(ck('alfa'), 'alfa')
        assert_equal(ck('bravot'), 'charlie')
        assert_equal(ck('charlie'), 'charlie')
        assert_equal(ck('kiloton'), 'lima')
        assert_equal(ck('oskar'), None)
        assert_equal(list(hr.keys()), keys)
        assert_equal(list(hr.values()), values)
        assert_equal(list(hr.keys_from('f')), keys[5:])
        hr.close()
    
def test_stored_fields():
    with TempStorage("storedfields") as st:
        sf = st.create_file("test.sf")
        sfw = StoredFieldWriter(sf, ["a", "b"])
        sfw.append({"a": "hello", "b": "there"})
        sfw.append({"a": "one", "b": "two"})
        sfw.append({"a": "alfa", "b": "bravo"})
        sfw.close()
        
        sf = st.open_file("test.sf")
        sfr = StoredFieldReader(sf)
        assert_equal(sfr[0], {"a": "hello", "b": "there"})
        assert_equal(sfr[2], {"a": "alfa", "b": "bravo"})
        assert_equal(sfr[1], {"a": "one", "b": "two"})
        sfr.close()
    
def test_termindex():
    terms = [("a", "alfa"), ("a", "bravo"), ("a", "charlie"), ("a", "delta"),
             ("b", "able"), ("b", "baker"), ("b", "dog"), ("b", "easy")]
    st = RamStorage()
    
    tw = TermIndexWriter(st.create_file("test.trm"))
    for i, t in enumerate(terms):
        tw.add(t, (1.0, i * 1000, 1))
    tw.close()
    
    tr = TermIndexReader(st.open_file("test.trm"))
    for i, (t1, t2) in enumerate(zip(tr.keys(), terms)):
        assert_equal(t1, t2)
        assert_equal(tr.get(t1), (1.0, i * 1000, 1))
        
    




from __future__ import with_statement

from nose.tools import assert_equal

from whoosh import analysis, fields, formats
from whoosh.support.testing import TempIndex


def test_vector_reading():
    a = analysis.StandardAnalyzer()
    schema = fields.Schema(title = fields.TEXT,
                           content = fields.TEXT(vector=formats.Frequency(analyzer=a)))
    
    with TempIndex(schema, "vectorreading") as ix:
        writer = ix.writer()
        writer.add_document(title=u"one",
                            content=u"This is the story of the black hole story")
        writer.commit()
        
        with ix.reader() as r:
            assert_equal(list(r.vector_as("frequency", 0, "content")),
                             [(u'black', 1), (u'hole', 1), (u'story', 2)])

def test_vector_merge():
    a = analysis.StandardAnalyzer()
    schema = fields.Schema(title = fields.TEXT,
                           content = fields.TEXT(vector=formats.Frequency(analyzer=a)))
    
    with TempIndex(schema, "vectormerge") as ix:
        writer = ix.writer()
        writer.add_document(title=u"one",
                            content=u"This is the story of the black hole story")
        writer.commit()
        
        writer = ix.writer()
        writer.add_document(title=u"two",
                            content=u"You can read along in your book")
        writer.commit()
        
        with ix.searcher() as s:
            r = s.reader()
        
            docnum = s.document_number(title=u"one")
            vec = list(r.vector_as("frequency", docnum, "content"))
            assert_equal(vec, [(u'black', 1), (u'hole', 1), (u'story', 2)])
        
            docnum = s.document_number(title=u"two")
        
            vec = list(r.vector_as("frequency", docnum, "content"))
            assert_equal(vec, [(u'along', 1), (u'book', 1), (u'read', 1)])
        
def test_vector_unicode():
    a = analysis.StandardAnalyzer()
    schema = fields.Schema(content = fields.TEXT(vector=formats.Frequency(analyzer=a)))
    
    with TempIndex(schema, "vectorunicode") as ix:
        writer = ix.writer()
        writer.add_document(content=u"\u1234\u2345\u3456 \u4567\u5678\u6789")
        writer.add_document(content=u"\u0123\u1234\u4567 \u4567\u5678\u6789")
        writer.commit()
        
        writer = ix.writer()
        writer.add_document(content=u"\u2345\u3456\u4567 \u789a\u789b\u789c")
        writer.add_document(content=u"\u0123\u1234\u4567 \u2345\u3456\u4567")
        writer.commit()
        
        with ix.reader() as r:
            vec = list(r.vector_as("frequency", 0, "content"))
            assert_equal(vec, [(u'\u3456\u4567', 1), (u'\u789a\u789b\u789c', 1)])





# coding=utf-8

from nose.tools import assert_equal

from whoosh import analysis, fields, qparser, query
from whoosh.qparser import dateparse


def test_empty_querystring():
    s = fields.Schema(content=fields.TEXT, title=fields.TEXT, id=fields.ID)
    qp = qparser.QueryParser("content", s)
    q = qp.parse(u"")
    assert_equal(q, query.NullQuery)

def test_fields():
    s = fields.Schema(content=fields.TEXT, title=fields.TEXT, id=fields.ID)
    qp = qparser.QueryParser("content", s)
    q = qp.parse(u"test")
    assert_equal(q.__class__, query.Term)
    assert_equal(q.fieldname, "content")
    assert_equal(q.text, "test")
    
    mq = qparser.MultifieldParser(("title", "content"), s)
    q = mq.parse(u"test")
    assert_equal(q.__class__, query.Or)
    assert_equal(q[0].__class__, query.Term)
    assert_equal(q[1].__class__, query.Term)
    assert_equal(q[0].fieldname, "title")
    assert_equal(q[1].fieldname, "content")
    assert_equal(q[0].text, "test")
    assert_equal(q[1].text, "test")
    
    q = mq.parse(u"title:test")
    assert_equal(q.__class__, query.Term)
    assert_equal(q.fieldname, "title")
    assert_equal(q.text, "test")

def test_multifield():
    schema = fields.Schema(content=fields.TEXT, title=fields.TEXT,
                           cat=fields.KEYWORD, date=fields.DATETIME)
    
    qs = u"time (cinema muza cat:place) OR (cinema muza cat:event)"
    qp = qparser.MultifieldParser(['content', 'title'], schema)
    q = qp.parse(qs)
    assert_equal(unicode(q), "((content:time OR title:time) AND (((content:cinema OR title:cinema) AND (content:muza OR title:muza) AND cat:place) OR ((content:cinema OR title:cinema) AND (content:muza OR title:muza) AND cat:event)))")
    
def test_fieldname_chars():
    s = fields.Schema(abc123=fields.TEXT, nisbah=fields.KEYWORD)
    qp = qparser.QueryParser("content", s)
    fieldmap = {'nisbah': [u'\u0646\u0633\u0628\u0629'],
                'abc123': ['xyz']}
    qp.add_plugin(qparser.FieldAliasPlugin(fieldmap))
    
    q = qp.parse(u"abc123:456")
    assert_equal(q.__class__, query.Term)
    assert_equal(q.fieldname, u'abc123')
    assert_equal(q.text, u'456')
    
    q = qp.parse(u"abc123:456 def")
    assert_equal(unicode(q), u"(abc123:456 AND content:def)")
    
    q = qp.parse(u'\u0646\u0633\u0628\u0629:\u0627\u0644\u0641\u0644\u0633\u0637\u064a\u0646\u064a')
    assert_equal(q.__class__, query.Term)
    assert_equal(q.fieldname, u'nisbah')
    assert_equal(q.text, u'\u0627\u0644\u0641\u0644\u0633\u0637\u064a\u0646\u064a')
    
    q = qp.parse(u"abc123 (xyz:123 OR qrs)")
    assert_equal(unicode(q), "(content:abc123 AND (abc123:123 OR content:qrs))")

def test_colonspace():
    s = fields.Schema(content=fields.TEXT, url=fields.ID)
    qp = qparser.QueryParser("content", s)
    q = qp.parse(u"url:test")
    assert_equal(q.__class__, query.Term)
    assert_equal(q.fieldname, "url")
    assert_equal(q.text, "test")
    
    q = qp.parse(u"url: test")
    assert q.__class__, query.And
    assert_equal(q[0].__class__, query.Term)
    assert_equal(q[1].__class__, query.Term)
    assert_equal(q[0].fieldname, "content")
    assert_equal(q[1].fieldname, "content")
    assert_equal(q[0].text, "url")
    assert_equal(q[1].text, "test")
    
    q = qp.parse(u"url:")
    assert_equal(q.__class__, query.Term)
    assert_equal(q.fieldname, "content")
    assert_equal(q.text, "url")
    
    s = fields.Schema(foo=fields.KEYWORD)
    qp = qparser.QueryParser("foo", s)
    q = qp.parse(u"blah:")
    assert_equal(q.__class__, query.Term)
    assert_equal(q.fieldname, "foo")
    assert_equal(q.text, "blah:")

def test_andor():
    qp  = qparser.QueryParser("a", None)
    q = qp.parse("a AND b OR c AND d OR e AND f")
    assert_equal(unicode(q), u"((a:a AND a:b) OR (a:c AND a:d) OR (a:e AND a:f))")
    
    q = qp.parse("aORb")
    assert_equal(q, query.Term("a", "aORb"))
    
    q = qp.parse("aOR b")
    assert_equal(q, query.And([query.Term("a", "aOR"), query.Term("a", "b")]))
    
    q = qp.parse("a ORb")
    assert_equal(q, query.And([query.Term("a", "a"), query.Term("a", "ORb")]))
    
    assert_equal(qp.parse("OR"), query.Term("a", "OR"))

def test_andnot():
    qp = qparser.QueryParser("content", None)
    q = qp.parse(u"this ANDNOT that")
    assert_equal(q.__class__, query.AndNot)
    assert_equal(q.a.__class__, query.Term)
    assert_equal(q.b.__class__, query.Term)
    assert_equal(q.a.text, "this")
    assert_equal(q.b.text, "that")
    
    q = qp.parse(u"foo ANDNOT bar baz")
    assert_equal(q.__class__, query.And)
    assert_equal(len(q), 2)
    assert_equal(q[0].__class__, query.AndNot)
    assert_equal(q[1].__class__, query.Term)
    
    q = qp.parse(u"foo fie ANDNOT bar baz")
    assert_equal(q.__class__, query.And)
    assert_equal(len(q), 3)
    assert_equal(q[0].__class__, query.Term)
    assert_equal(q[1].__class__, query.AndNot)
    assert_equal(q[2].__class__, query.Term)
    
    q = qp.parse(u"a AND b ANDNOT c")
    assert_equal(q.__class__, query.AndNot)
    assert_equal(unicode(q), "((content:a AND content:b) ANDNOT content:c)")

def test_boost():
    qp = qparser.QueryParser("content", None)
    q = qp.parse(u"this^3 fn:that^0.5 5.67 hi^5x")
    assert_equal(q[0].boost, 3.0)
    assert_equal(q[1].boost, 0.5)
    assert_equal(q[1].fieldname, "fn")
    assert_equal(q[2].text, "5.67")
    assert_equal(q[3].text, "hi^5x")
    
    q = qp.parse("alfa (bravo OR charlie)^2.5 ^3")
    assert_equal(len(q), 3)
    assert_equal(q[0].boost, 1.0)
    assert_equal(q[1].boost, 2.5)
    assert_equal(q[2].text, "^3")
    
def test_wildcard1():
    qp = qparser.QueryParser("content", None)
    q = qp.parse(u"hello *the?e* ?star*s? test")
    assert_equal(len(q), 4)
    assert_equal(q[0].__class__, query.Term)
    assert_equal(q[0].text, "hello")
    assert_equal(q[1].__class__, query.Wildcard)
    assert_equal(q[1].text, "*the?e*")
    assert_equal(q[2].__class__, query.Wildcard)
    assert_equal(q[2].text, "?star*s?")
    assert_equal(q[3].__class__, query.Term)
    assert_equal(q[3].text, "test")
    
def test_wildcard2():
    qp = qparser.QueryParser("content", None)
    q = qp.parse(u"*the?e*")
    assert_equal(q.__class__, query.Wildcard)
    assert_equal(q.text, "*the?e*")
    
def test_parse_fieldname_underscores():
    s = fields.Schema(my_name=fields.ID(stored=True), my_value=fields.TEXT)
    qp = qparser.QueryParser("my_value", schema=s)
    q = qp.parse(u"my_name:Green")
    assert_equal(q.__class__, query.Term)
    assert_equal(q.fieldname, "my_name")
    assert_equal(q.text, "Green")

def test_endstar():
    qp = qparser.QueryParser("text", None)
    q = qp.parse(u"word*")
    assert_equal(q.__class__, query.Prefix)
    assert_equal(q.text, "word")
    
    q = qp.parse(u"first* second")
    assert_equal(q[0].__class__, query.Prefix)
    assert_equal(q[0].text, "first")

def test_singlequotes():
    qp = qparser.QueryParser("text", None)
    q = qp.parse("hell's hot 'i stab at thee'")
    assert_equal(q.__class__.__name__, 'And')
    assert_equal(len(q), 3)
    assert_equal(q[0].__class__, query.Term)
    assert_equal(q[1].__class__, query.Term)
    assert_equal(q[2].__class__, query.Term)
    assert_equal(q[0].text, "hell's")
    assert_equal(q[1].text, "hot")
    assert_equal(q[2].text, "i stab at thee")
    
    q = qp.parse("alfa zulu:'bravo charlie' delta")
    assert_equal(q.__class__.__name__, 'And')
    assert_equal(len(q), 3)
    assert_equal(q[0].__class__, query.Term)
    assert_equal(q[1].__class__, query.Term)
    assert_equal(q[2].__class__, query.Term)
    assert_equal((q[0].fieldname, q[0].text), ("text", "alfa"))
    assert_equal((q[1].fieldname, q[1].text), ("zulu", "bravo charlie"))
    assert_equal((q[2].fieldname, q[2].text), ("text", "delta"))
    
    q = qp.parse("The rest 'is silence")
    assert_equal(q.__class__, query.And)
    assert_equal(len(q), 4)
    assert_equal([t.text for t in q.subqueries], ["The", "rest", "'is" ,"silence"])
    
    q = qp.parse("I don't like W's stupid face")
    assert_equal(q.__class__, query.And)
    assert_equal(len(q), 6)
    assert_equal([t.text for t in q.subqueries], ["I", "don't", "like" ,"W's", "stupid", "face"])
    
    q = qp.parse("I forgot the drinkin' in '98")
    assert_equal(q.__class__, query.And)
    assert_equal(len(q), 6)
    assert_equal([t.text for t in q.subqueries], ["I", "forgot", "the" ,"drinkin'", "in", "'98"])

#    def test_escaping():
#        qp = qparser.QueryParser("text", None)
#        
#        q = qp.parse(r'big\small')
#        assert q.__class__, query.Term, q)
#        assert q.text, "bigsmall")
#        
#        q = qp.parse(r'big\\small')
#        assert q.__class__, query.Term)
#        assert q.text, r'big\small')
#        
#        q = qp.parse(r'http\:example')
#        assert q.__class__, query.Term)
#        assert q.fieldname, "text")
#        assert q.text, "http:example")
#        
#        q = qp.parse(r'hello\ there')
#        assert q.__class__, query.Term)
#        assert q.text, "hello there")
#        
#        q = qp.parse(r'\[start\ TO\ end\]')
#        assert q.__class__, query.Term)
#        assert q.text, "[start TO end]")
#    
#        schema = fields.Schema(text=fields.TEXT)
#        qp = qparser.QueryParser("text", None)
#        q = qp.parse(r"http\:\/\/www\.example\.com")
#        assert q.__class__, query.Term)
#        assert q.text, "http://www.example.com")
#        
#        q = qp.parse(u"\\\\")
#        assert q.__class__, query.Term)
#        assert q.text, "\\")

#    def test_escaping_wildcards():
#        qp = qparser.QueryParser("text", None)
#        
#        q = qp.parse(u"a*b*c?d")
#        assert q.__class__, query.Wildcard)
#        assert q.text, "a*b*c?d")
#        
#        q = qp.parse(u"a*b\\*c?d")
#        assert q.__class__, query.Wildcard)
#        assert q.text, "a*b*c?d")
#        
#        q = qp.parse(u"a*b\\\\*c?d")
#        assert q.__class__, query.Wildcard)
#        assert q.text, u'a*b\\*c?d')
#        
#        q = qp.parse(u"ab*")
#        assert q.__class__, query.Prefix)
#        assert q.text, u"ab")
#        
#        q = qp.parse(u"ab\\\\*")
#        assert q.__class__, query.Wildcard)
#        assert q.text, u"ab\\*")
    
def test_phrase():
    qp = qparser.QueryParser("content", None)
    q = qp.parse('"alfa bravo" "charlie delta echo"^2.2 test:"foxtrot golf"')
    assert_equal(q[0].__class__, query.Phrase)
    assert_equal(q[0].words, ["alfa", "bravo"])
    assert_equal(q[1].__class__, query.Phrase)
    assert_equal(q[1].words, ["charlie", "delta", "echo"])
    assert_equal(q[1].boost, 2.2)
    assert_equal(q[2].__class__, query.Phrase)
    assert_equal(q[2].words, ["foxtrot", "golf"])
    assert_equal(q[2].fieldname, "test")
    
def test_weird_characters():
    qp = qparser.QueryParser("content", None)
    q = qp.parse(u".abcd@gmail.com")
    assert_equal(q.__class__, query.Term)
    assert_equal(q.text, u".abcd@gmail.com")
    q = qp.parse(u"r*")
    assert_equal(q.__class__, query.Prefix)
    assert_equal(q.text, u"r")
    q = qp.parse(u".")
    assert_equal(q.__class__, query.Term)
    assert_equal(q.text, u".")
    q = qp.parse(u"?")
    assert_equal(q.__class__, query.Wildcard)
    assert_equal(q.text, u"?")
    
def test_euro_chars():
    schema = fields.Schema(text=fields.TEXT)
    qp = qparser.QueryParser("text", schema)
    q = qp.parse(u"straße")
    assert_equal(q.__class__, query.Term)
    assert_equal(q.text, u"straße")

def test_star():
    schema = fields.Schema(text = fields.TEXT(stored=True))
    qp = qparser.QueryParser("text", schema)
    q = qp.parse(u"*")
    assert_equal(q.__class__, query.Every)

    q = qp.parse(u"*h?ll*")
    assert_equal(q.__class__, query.Wildcard)
    assert_equal(q.text, u"*h?ll*")
    
    q = qp.parse(u"h?pe")
    assert_equal(q.__class__, query.Wildcard)
    assert_equal(q.text, u"h?pe")
    
    q = qp.parse(u"*? blah")
    assert_equal(q.__class__, query.And)
    assert_equal(q[0].__class__, query.Wildcard)
    assert_equal(q[0].text, u"*?")
    assert_equal(q[1].__class__, query.Term)
    assert_equal(q[1].text, u"blah")
    
    q = qp.parse(u"*ending")
    assert_equal(q.__class__, query.Wildcard)
    assert_equal(q.text, u"*ending")
    
    q = qp.parse(u"*q")
    assert_equal(q.__class__, query.Wildcard)
    assert_equal(q.text, u"*q")

def test_range():
    schema = fields.Schema(name=fields.ID(stored=True), text = fields.TEXT(stored=True))
    qp = qparser.QueryParser("text", schema)
    
    q = qp.parse(u"[alfa to bravo}")
    assert_equal(q.__class__, query.TermRange)
    assert_equal(q.start, "alfa")
    assert_equal(q.end, "bravo")
    assert_equal(q.startexcl, False)
    assert_equal(q.endexcl, True)
    
    q = qp.parse(u"['hello there' to 'what ever']")
    assert_equal(q.__class__, query.TermRange)
    assert_equal(q.start, "hello there")
    assert_equal(q.end, "what ever")
    assert_equal(q.startexcl, False)
    assert_equal(q.endexcl, False)
    
    q = qp.parse(u"name:{'to' to 'b'}")
    assert_equal(q.__class__, query.TermRange)
    assert_equal(q.start, "to")
    assert_equal(q.end, "b")
    assert_equal(q.startexcl, True)
    assert_equal(q.endexcl, True)
    
    q = qp.parse(u"name:{'a' to 'to']")
    assert_equal(q.__class__, query.TermRange)
    assert_equal(q.start, "a")
    assert_equal(q.end, "to")
    assert_equal(q.startexcl, True)
    assert_equal(q.endexcl, False)
    
    q = qp.parse(u"name:[a to to]")
    assert_equal(q.__class__, query.TermRange)
    assert_equal(q.start, "a")
    assert_equal(q.end, "to")
    
    q = qp.parse(u"name:[to to b]")
    assert_equal(q.__class__, query.TermRange)
    assert_equal(q.start, "to")
    assert_equal(q.end, "b")
    
    q = qp.parse(u"[alfa to alfa]")
    assert_equal(q.__class__, query.Term)
    assert_equal(q.text, "alfa")
    
    q = qp.parse(u"Ind* AND name:[d TO]")
    assert_equal(q.__class__, query.And)
    assert_equal(q[0].__class__, query.Prefix)
    assert_equal(q[1].__class__, query.TermRange)
    assert_equal(q[0].text, "ind")
    assert_equal(q[1].start, "d")
    assert_equal(q[1].fieldname, "name")
    
    q = qp.parse(u"name:[d TO]")
    assert_equal(q.__class__, query.TermRange)
    assert_equal(q.start, "d")
    assert_equal(q.fieldname, "name")

def test_numeric_range():
    schema = fields.Schema(id=fields.STORED, number=fields.NUMERIC)
    qp = qparser.QueryParser("number", schema)
    
    teststart = 40
    testend = 100
    
    q = qp.parse("[%s to *]" % teststart)
    assert_equal(q, query.NullQuery)
    
    q = qp.parse("[%s to]" % teststart)
    assert_equal(q.__class__, query.NumericRange)
    assert_equal(q.start, teststart)
    assert_equal(q.end, None)
    
    q = qp.parse("[to %s]" % testend)
    assert_equal(q.__class__, query.NumericRange)
    assert_equal(q.start, None)
    assert_equal(q.end, testend)
    
    q = qp.parse("[%s to %s]" % (teststart, testend))
    assert_equal(q.__class__, query.NumericRange)
    assert_equal(q.start, teststart)
    assert_equal(q.end, testend)
    
def test_regressions():
    qp = qparser.QueryParser("f", None)
    
    # From 0.3.18, these used to require escaping. Mostly good for
    # regression testing.
    assert_equal(qp.parse(u"re-inker"), query.Term("f", u"re-inker"))
    assert_equal(qp.parse(u"0.7 wire"), query.And([query.Term("f", u"0.7"), query.Term("f", u"wire")]))
    assert (qp.parse(u"daler-rowney pearl 'bell bronze'")
            == query.And([query.Term("f", u"daler-rowney"),
                          query.Term("f", u"pearl"),
                          query.Term("f", u"bell bronze")]))
    
    q = qp.parse(u'22" BX')
    assert_equal(q, query.And([query.Term("f", u'22"'), query.Term("f", "BX")]))
    
def test_empty_ranges():
    schema = fields.Schema(name=fields.TEXT, num=fields.NUMERIC,
                           date=fields.DATETIME)
    qp = qparser.QueryParser("text", schema)
    
    for fname in ("name", "date"):
        q = qp.parse("%s:[to]" % fname)
        assert_equal(q.__class__, query.Every)

def test_empty_numeric_range():
    schema = fields.Schema(id=fields.ID, num=fields.NUMERIC)
    qp = qparser.QueryParser("num", schema)
    q = qp.parse("num:[to]")
    assert_equal(q.__class__, query.NumericRange)
    assert_equal(q.start, None)
    assert_equal(q.end, None)

def test_nonexistant_fieldnames():
    # Need an analyzer that won't mangle a URL
    a = analysis.SimpleAnalyzer("\\S+")
    schema = fields.Schema(id=fields.ID, text=fields.TEXT(analyzer=a))
    
    qp = qparser.QueryParser("text", schema)
    q = qp.parse(u"id:/code http://localhost/")
    assert_equal(q.__class__, query.And)
    assert_equal(q[0].__class__, query.Term)
    assert_equal(q[0].fieldname, "id")
    assert_equal(q[0].text, "/code")
    assert_equal(q[1].__class__, query.Term)
    assert_equal(q[1].fieldname, "text")
    assert_equal(q[1].text, "http://localhost/")

def test_stopped():
    schema = fields.Schema(text = fields.TEXT)
    qp = qparser.QueryParser("text", schema)
    q = qp.parse(u"a b")
    assert_equal(q, query.NullQuery)
    
def test_analyzing_terms():
    schema = fields.Schema(text=fields.TEXT(analyzer=analysis.StemmingAnalyzer()))
    qp = qparser.QueryParser("text", schema)
    q = qp.parse(u"Indexed!")
    assert_equal(q.__class__, query.Term)
    assert_equal(q.text, "index")
    
def test_simple():
    parser = qparser.SimpleParser("x", None)
    q = parser.parse(u"alfa bravo charlie delta")
    assert_equal(unicode(q), u"(x:alfa OR x:bravo OR x:charlie OR x:delta)")
    
    q = parser.parse(u"alfa +bravo charlie delta")
    assert_equal(unicode(q), u"(x:bravo ANDMAYBE (x:alfa OR x:charlie OR x:delta))")
    
    q = parser.parse(u"alfa +bravo -charlie delta")
    assert_equal(unicode(q), u"((x:bravo ANDMAYBE (x:alfa OR x:delta)) ANDNOT x:charlie)")
    
    q = parser.parse(u"- alfa +bravo + delta")
    assert_equal(unicode(q), u"((x:bravo AND x:delta) ANDNOT x:alfa)")

def test_dismax():
    parser = qparser.DisMaxParser({"body": 0.8, "title": 2.5}, None)
    q = parser.parse(u"alfa bravo charlie")
    assert_equal(unicode(q), u"(DisMax(body:alfa^0.8 title:alfa^2.5) OR DisMax(body:bravo^0.8 title:bravo^2.5) OR DisMax(body:charlie^0.8 title:charlie^2.5))")
    
    q = parser.parse(u"alfa +bravo charlie")
    assert_equal(unicode(q), u"(DisMax(body:bravo^0.8 title:bravo^2.5) ANDMAYBE (DisMax(body:alfa^0.8 title:alfa^2.5) OR DisMax(body:charlie^0.8 title:charlie^2.5)))")
    
    q = parser.parse(u"alfa -bravo charlie")
    assert_equal(unicode(q), u"((DisMax(body:alfa^0.8 title:alfa^2.5) OR DisMax(body:charlie^0.8 title:charlie^2.5)) ANDNOT DisMax(body:bravo^0.8 title:bravo^2.5))")
    
    q = parser.parse(u"alfa -bravo +charlie")
    assert_equal(unicode(q), u"((DisMax(body:charlie^0.8 title:charlie^2.5) ANDMAYBE DisMax(body:alfa^0.8 title:alfa^2.5)) ANDNOT DisMax(body:bravo^0.8 title:bravo^2.5))")

def test_many_clauses():
    qs = "1" + (" OR 1" * 1000)
    
    parser = qparser.QueryParser("content", None)
    parser.parse(qs)
    
def test_roundtrip():
    parser = qparser.QueryParser("a", None)
    q = parser.parse(u"a OR ((b AND c AND d AND e) OR f OR g) ANDNOT h")
    assert_equal(unicode(q), u"((a:a OR (a:b AND a:c AND a:d AND a:e) OR a:f OR a:g) ANDNOT a:h)")
    
def test_ngrams():
    schema = fields.Schema(grams=fields.NGRAM)
    parser = qparser.QueryParser('grams', schema)
    parser.remove_plugin_class(qparser.WhitespacePlugin)
    
    q = parser.parse(u"Hello There")
    assert_equal(q.__class__, query.And)
    assert_equal(len(q), 8)
    assert_equal([sq.text for sq in q], ["hell", "ello", "llo ", "lo t", "o th", " the", "ther", "here"])
    
def test_ngramwords():
    schema = fields.Schema(grams=fields.NGRAMWORDS(queryor=True))
    parser = qparser.QueryParser('grams', schema)
    
    q = parser.parse(u"Hello Tom")
    assert_equal(q.__class__, query.And)
    assert_equal(q[0].__class__, query.Or)
    assert_equal(q[1].__class__, query.Term)
    assert_equal(q[0][0].text, "hell")
    assert_equal(q[0][1].text, "ello")
    assert_equal(q[1].text, "tom")

def test_multitoken_words():
    textfield = fields.TEXT()
    textfield.multitoken_query = "or"
    schema = fields.Schema(text=textfield)
    parser = qparser.QueryParser('text', schema)
    qstring = u"chaw-bacon"
    
    texts = list(schema["text"].process_text(qstring))
    assert_equal(texts, ["chaw", "bacon"])
    
    q = parser.parse(qstring)
    assert_equal(q.__class__, query.Or)
    assert_equal(len(q), 2)
    assert_equal(q[0].__class__, query.Term)
    assert_equal(q[0].text, "chaw")
    assert_equal(q[1].__class__, query.Term)
    assert_equal(q[1].text, "bacon")

def test_operators():
    qp = qparser.QueryParser("f", None)
    
    q = qp.parse("a AND b OR c AND d")
    assert_equal(unicode(q), "((f:a AND f:b) OR (f:c AND f:d))")
    
    q = qp.parse("a OR b OR c OR d")
    assert_equal(unicode(q), "(f:a OR f:b OR f:c OR f:d)")
    
    q = qp.parse("a ANDMAYBE b ANDNOT c REQUIRE d")
    assert_equal(unicode(q), "((f:a ANDMAYBE (f:b ANDNOT f:c)) REQUIRE f:d)")

def test_associativity():
    left_andmaybe = (qparser.InfixOperator("ANDMAYBE", qparser.AndMaybeGroup, True), 0)
    right_andmaybe = (qparser.InfixOperator("ANDMAYBE", qparser.AndMaybeGroup, False), 0)
    not_ = (qparser.PrefixOperator("NOT", qparser.NotGroup), 0)
    
    def make_parser(*ops):
        parser = qparser.QueryParser("f", None)
        parser.replace_plugin(qparser.CompoundsPlugin(ops, clean=True))
        return parser
    
    p = make_parser(left_andmaybe)
    q = p.parse("a ANDMAYBE b ANDMAYBE c ANDMAYBE d")
    assert_equal(unicode(q), "(((f:a ANDMAYBE f:b) ANDMAYBE f:c) ANDMAYBE f:d)")
    
    p = make_parser(right_andmaybe)
    q = p.parse("a ANDMAYBE b ANDMAYBE c ANDMAYBE d")
    assert_equal(unicode(q), "(f:a ANDMAYBE (f:b ANDMAYBE (f:c ANDMAYBE f:d)))")
    
    p = make_parser(not_)
    q = p.parse("a NOT b NOT c NOT d", normalize=False)
    assert_equal(unicode(q), "(f:a AND NOT f:b AND NOT f:c AND NOT f:d)")
    
    p = make_parser(left_andmaybe)
    q = p.parse("(a ANDMAYBE b) ANDMAYBE (c ANDMAYBE d)")
    assert_equal(unicode(q), "((f:a ANDMAYBE f:b) ANDMAYBE (f:c ANDMAYBE f:d))")
    
    p = make_parser(right_andmaybe)
    q = p.parse("(a ANDMAYBE b) ANDMAYBE (c ANDMAYBE d)")
    assert_equal(unicode(q), "((f:a ANDMAYBE f:b) ANDMAYBE (f:c ANDMAYBE f:d))")
    
def test_not_assoc():
    qp = qparser.QueryParser("text", None)
    q = qp.parse(u"a AND NOT b OR c")
    assert_equal(unicode(q), "((text:a AND NOT text:b) OR text:c)")
    
    qp = qparser.QueryParser("text", None)
    q = qp.parse(u"a NOT (b OR c)")
    assert_equal(unicode(q), "(text:a AND NOT (text:b OR text:c))")
    
def test_fieldname_space():
    qp = qparser.QueryParser("a", None)
    q = qp.parse("Man Ray: a retrospective")
    assert_equal(unicode(q), "(a:Man AND a:Ray: AND a:a AND a:retrospective)")
    
def test_fieldname_fieldname():
    qp = qparser.QueryParser("a", None)
    q = qp.parse("a:b:")
    assert_equal(q, query.Term("a", u"b:"))
    
def test_paren_fieldname():
    schema = fields.Schema(kind=fields.ID, content=fields.TEXT) 

    qp = qparser.QueryParser("content", schema)
    q = qp.parse(u"(kind:1d565 OR kind:7c584) AND (stuff)")
    assert_equal(unicode(q), "((kind:1d565 OR kind:7c584) AND content:stuff)")

    q = qp.parse(u"kind:(1d565 OR 7c584) AND (stuff)")
    assert_equal(unicode(q), "((kind:1d565 OR kind:7c584) AND content:stuff)")

def test_star_paren():
    qp = qparser.QueryParser("content", None)
    q = qp.parse(u"(*john*) AND (title:blog)")
    
    assert_equal(q.__class__, query.And)
    assert_equal(q[0].__class__, query.Wildcard)
    assert_equal(q[1].__class__, query.Term)
    assert_equal(q[0].fieldname, "content")
    assert_equal(q[1].fieldname, "title")
    assert_equal(q[0].text, "*john*")
    assert_equal(q[1].text, "blog")

    





